package checkers.basetype;

import java.io.IOException;
import java.util.Properties;

import com.sun.source.tree.CompilationUnitTree;

import checkers.source.SourceChecker;
import checkers.types.*;

import javax.lang.model.*;
import javax.annotation.processing.*;

/**
 * An abstract {@link SourceChecker} that provides a simple {@link
 * checkers.source.SourceVisitor} implementation for typical assignment and
 * pseudo-assignment checking of annotated types. Most typechecker plug-ins will
 * want to extend this class.
 * 
 * <p>
 * 
 * Subclasses must implement the {@link BaseTypeChecker#isSubtype} method to
 * define the relationships between various annotated and unannotated types in
 * their custom type systems. Note that {@link BaseTypeChecker#isSubtype} must
 * operate on type annotations and not the underlying types (which are checked
 * by the compiler's built-in type checker).
 * 
 * @see BaseTypeChecker#isSubtype(AnnotatedTypeMirror, AnnotatedTypeMirror)
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes({"*"})
public abstract class BaseTypeChecker extends SourceChecker {

    @Override
    protected Properties getMessages(String fileName) throws IOException {
        return new Properties();
    }

    @Override
    protected BaseTypeVisitor<?, ?> getSourceVisitor(CompilationUnitTree root) {
        return new BaseTypeVisitor<Void, Void>(this, root);
    }

    /**
     * Tests whether the one annotated type is a subtype of another, with
     * respect to the annotations on these types. Subclasses may wish to ignore
     * annotations that are not related to the type qualifiers they check.
     * 
     * @param sup the parent type
     * @param sub the child type
     * @return true iff {@code sub} is a subtype of {@code sup}
     */
    public abstract boolean isSubtype(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub);

    /**
     * Tests whether the one annotated type is a subtype of another, with
     * respect to the annotations on these types, without descending into
     * generic or array types (i.e., only performing the subtype check on the
     * raw type or outermost array dimension). Subclasses may wish to ignore
     * annotations that are not related to the type qualifiers they check.
     * 
     * @see BaseTypeChecker#isSubtype(AnnotatedTypeMirror, AnnotatedTypeMirror)
     * @param sup the parent type
     * @param sub the child type
     * @return true iff {@code sub} is a subtype of {@code sup}
     * @throws UnsupportedOperationException if this method is not overriden
     */
    public boolean isSubtypeIgnoringTypeParameters(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub) {
        throw new java.lang.UnsupportedOperationException("isSubtypeIgnoringTypeParameters not implemented on this checker");
    }

    // TODO: eventually, make isSubtype non-abstract and dependent on isSubtypeIgnoringTypeParameters?
    
}
