package checkers.basetype;

import java.util.List;
import java.util.Map;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.Name;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.DeclaredType;

import com.sun.source.tree.*;
import com.sun.source.util.SourcePositions;
import com.sun.source.util.TreePath;
import com.sun.source.util.Trees;

import checkers.nonnull.NonNullChecker;
import checkers.source.*;
import checkers.types.*;
import checkers.util.*;
import static checkers.types.AnnotatedTypeMirror.*;

/**
 * A {@link SourceVisitor} that performs assignment and pseudo-assignment
 * checking.
 * 
 * <p>
 * 
 * This implementation uses the {@link AnnotatedTypeFactory} implementation
 * provided by an associated {@link BaseTypeChecker}; its visitor methods will
 * invoke this factory on parts of the AST to determine the "annotated type" of
 * an expression. Then, the visitor methods will check the types in assignments
 * and pseudo-assignments using {@link #commonAssignmentCheck}, which
 * ultimately calls the {@link BaseTypeChecker#isSubtype} method and reports
 * errors that violate Java's rules of assignment.
 * 
 * <p>
 * 
 * Note that since this implementation only performs assignment and
 * pseudo-assignment checking, other rules for custom type systems must be added
 * in subclasses (e.g., dereference checking in the {@link NonNullChecker} is
 * implemented in the {@link NonNullChecker}'s
 * {@link com.sun.source.util.TreeScanner#visitMemberSelect} method).
 * 
 * @see JLS $4
 * @see BaseTypeChecker#isSubtype(AnnotatedTypeMirror, AnnotatedTypeMirror)
 * @see AnnotatedTypeFactory
 */
public class BaseTypeVisitor<R, P> extends SourceVisitor<R, P> {

    /** The checker corresponding to this visitor. */
    protected BaseTypeChecker checker;

    /** The annotation factory to use for creating annotations. */
    protected final AnnotationFactory annoFactory;

    /** Type utilities, used for determining overriding methods. */
    private final TypesUtils typeUtils;
    
    /** The options that were provided to the checker using this visitor. */
    private final Map<String, String> options;

    /** For obtaining line numbers in -Ashowchecks debugging output. */
    private final SourcePositions positions;


    /**
     * @param checker the typechecker associated with this visitor (for
     *        callbacks to {@link BaseTypeChecker#isSubtype})
     * @param root the root of the AST that this visitor operates on
     */
    public BaseTypeVisitor(BaseTypeChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.checker = checker;
        
        ProcessingEnvironment env = checker.getProcessingEnvironment();
        this.annoFactory = new AnnotationFactory(env);
        this.typeUtils = new TypesUtils(env);
        this.options = env.getOptions();
        this.positions = trees.getSourcePositions();
    }
    
    @Override
    public R visitAnnotation(AnnotationTree node, P p) {
        // Skip checking inside annotations.
        return null;
    }

    @Override
    public R visitAssignment(AssignmentTree node, P p) {
        commonAssignmentCheck(node.getVariable(), node.getExpression(), "assignment.invalid", p);
        return super.visitAssignment(node, p);
    }

    @Override
    public R visitCompoundAssignment(CompoundAssignmentTree node, P p) {
        commonAssignmentCheck(node.getVariable(), node.getExpression(), "compound.assignment.invalid", p);
        return super.visitCompoundAssignment(node, p);
    }

    @Override
    public R visitEnhancedForLoop(EnhancedForLoopTree node, P p) {
        // TODO: Finish implementing enhanced for loop checking.
//        final AnnotatedTypeMirror var = null; // TODO: Get variable type
//        final AnnotatedTypeMirror iterableType = null;  // TODO: Get iterabletype
//        final AnnotatedTypeMirror iteratedType =
//            annoTypes.getIteratedType(iterableType);
//        commonAssignmentCheck(var, iteratedType, node.getExpression(), p);
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    public R visitMethod(MethodTree node, P p) {
        
        Element elt = InternalUtils.symbol(node);
        assert elt != null : "no symbol for method";
        
        // TODO: this currently checks for *any* receiver annotation, which is too conservative
        if (elt.getModifiers().contains(Modifier.STATIC) &&
                !node.getReceiver().getAnnotations().isEmpty())
            checker.report(Result.failure("missing.this"), node.getReceiver());
        
        // TODO: Finish implementing method override checks.

        // Find which method this overrides!        
        ExecutableElement methodElement = (ExecutableElement) InternalUtils.symbol(node);
        AnnotatedDeclaredType enclosingType = (AnnotatedDeclaredType)factory.getAnnotatedType(methodElement.getEnclosingElement());
        for (Map.Entry<DeclaredType, ExecutableElement> pair: typeUtils.overriddenMethods(methodElement).entrySet()) {
            AnnotatedExecutableType overriddenMethod = (AnnotatedExecutableType)factory.getAnnotatedType(pair.getValue());
            AnnotatedDeclaredType overriddenType = (AnnotatedDeclaredType)factory.getAnnotatedType(pair.getKey().asElement());
            checkOverride(node, enclosingType, overriddenMethod, overriddenType, p);
        }
        return super.visitMethod(node, p);
    }

    // TODO: move me to a utility class
    /**
     * @param node the method invocation to check
     * @return true if this is a super call to the {@link Enum} constructor
     */
    private boolean isEnumSuper(MethodInvocationTree node) {
        
        AnnotatedExecutableType invokedMethod = 
            (AnnotatedExecutableType)factory.getAnnotatedType(node.getMethodSelect());
        
        assert invokedMethod.getElement() instanceof ExecutableElement;
        
        ExecutableElement ex = (ExecutableElement)invokedMethod.getElement();
        
        Element sup = ex.getEnclosingElement();
        if (sup instanceof TypeElement &&
                ((TypeElement)sup).getQualifiedName().contentEquals("java.lang.Enum"))
            return true;
        
        return false;
    }
    
    @Override
    public R visitMethodInvocation(MethodInvocationTree node, P p) {
        if (InternalUtils.isAnonymousConstructor(getCurrentPath()))
            return super.visitMethodInvocation(node, p);

        AnnotatedExecutableType invokedMethod = (AnnotatedExecutableType)factory.getAnnotatedType(node.getMethodSelect());
        List<AnnotatedTypeMirror> requiredArgs = invokedMethod.getParameterTypes();
        
        // Skip calls to the Enum constructor (they're generated by javac and
        // hard to check).
        if (isEnumSuper(node))
            return super.visitMethodInvocation(node, p);
        
        // FIXME: check varargs arguments (this only checks up to varargs)
        
        int maxCheck = (invokedMethod.isVarArgs() ? requiredArgs.size() - 1 : requiredArgs.size());
        
        assert node.getArguments().size() >= maxCheck :
            String.format("not enough args for %s", node);
        
        for (int i = 0; i < maxCheck; ++i)
            commonAssignmentCheck(requiredArgs.get(i),
                    node.getArguments().get(i), 
                    "argument.invalid", p);
        
        return super.visitMethodInvocation(node, p);
    }

    @Override
    public R visitReturn(ReturnTree node, P p) {
        
        // Don't try to check return expressions for void methods.
        if (node.getExpression() == null)
            return super.visitReturn(node, p);

        MethodTree enclosingMethod =
            TreeUtils.enclosingMethod(getCurrentPath());
        
        AnnotatedExecutableType methodType = (AnnotatedExecutableType) factory.getAnnotatedType(enclosingMethod);
        commonAssignmentCheck(methodType.getReturnType(), node.getExpression(),
                "return.invalid", p);
        
        return super.visitReturn(node, p);
    }

    @Override
    public R visitVariable(VariableTree node, P p) {
        
        // If there's no assignment in this variable declaration, skip it.
        if (node.getInitializer() == null)
            return super.visitVariable(node, p);
        
        commonAssignmentCheck(node, node.getInitializer(), "assignment.invalid", p);
        return super.visitVariable(node, p);
    }

    // TODO: this should be moved to a utility class
    /**
     * Returns the receiver type of the expression, mainly for method
     * invocation and field accesses.
     */
    // TODO: Doesn't handle static method invocation
    protected AnnotatedDeclaredType getReceiver(ExpressionTree expression) {
        
        if (!(expression instanceof MethodInvocationTree
                || expression instanceof MemberSelectTree
                || expression instanceof IdentifierTree))
            // No receiver type for those
            return null;
        
    	Tree receiver = expression;
    	
    	if (receiver instanceof MethodInvocationTree)
    		receiver = ((MethodInvocationTree)receiver).getMethodSelect();

    	// Figure out if this a self type
    	if (receiver instanceof IdentifierTree)
    		return getSelfType();

    	if (receiver instanceof MemberSelectTree)
    		receiver = ((MemberSelectTree)receiver).getExpression();

    	if (receiver instanceof IdentifierTree) {
    		IdentifierTree iden = (IdentifierTree)receiver;
    		if (iden.getName().contentEquals("this")
    				|| iden.getName().contentEquals("super"))
    			return getSelfType();
    	}
    	return (AnnotatedDeclaredType) factory.getAnnotatedType(receiver);
    }
    
    /**
     * Returns the type of {@code this} in the current location, which can
     * be used if {@code this} has a special semantics (e.g. {@code this}
     * is non-null)
     */
    protected AnnotatedDeclaredType getSelfType() {
        // For now simply return the type of the enclosing class
        ClassTree tree = TreeUtils.enclosingClass(getCurrentPath());
        AnnotatedTypeMirror type = factory.getAnnotatedType(tree);
        assert type instanceof AnnotatedDeclaredType;
        return (AnnotatedDeclaredType)type;
    }
    
    /**
     * Checks that an overriding method's return type, parameter types, and
     * receiver type are correct with respect to the annotations on the
     * overridden method's return type, parameter types, and receiver type.
     * 
     * <p>
     * 
     * This method returns the result of the check, but also emits error
     * messages as a side effect.
     * 
     * @param overriderTree the AST node of the overriding method
     * @param enclosingType the declared type enclosing the overrider method
     * @param overriddenMethod the type of the overridden method
     * @param overriddenType the declared type enclosing the overridden method
     * @param p an optional parameter (as supplied to visitor methods)
     * @return true if the override check passed, false otherwise
     */
    protected boolean checkOverride(MethodTree overriderTree,
            AnnotatedDeclaredType enclosingType,
            AnnotatedExecutableType overriddenMethod,
            AnnotatedDeclaredType overriddenType,
            P p) {

        // Get the type of the overriding method.
        AnnotatedExecutableType overrider =
            (AnnotatedExecutableType) factory.getAnnotatedType(overriderTree);
        
        boolean result = true;
        
        // Check the return value.
        if (!checker.isSubtype(overriddenMethod.getReturnType(),
                overrider.getReturnType())) {
            checker.report(Result.failure("override.return.invalid",
                    overrider.toString(), enclosingType.toString(),
                    overriddenMethod.toString(), overriddenType.toString()),
                    overriderTree.getReturnType());
            // emit error message
            result = false;
        }

        // Check parameter values. (FIXME varargs)
        List<AnnotatedTypeMirror> overriderParams =
            overrider.getParameterTypes();
        List<AnnotatedTypeMirror> overriddenParams =
            overriddenMethod.getParameterTypes();
        for (int i = 0; i < overriderParams.size(); ++i) {
            if (!checker.isSubtype(overriderParams.get(i), overriddenParams.get(i))) {
                checker.report(Result.failure("override.param.invalid",
                        overrider.toString(),
                        enclosingType.toString(),
                        overriddenMethod.toString(),
                        overriddenType.toString(),
                        overriddenParams.get(i).toString(),
                        overriderParams.get(i).toString()
                        ), overriderTree.getParameters().get(i));
                // emit error message
                result = false;
            }
        }

        // Check the receiver type.
        if (!checker.isSubtype(overriddenMethod.getReceiverType(),
                overrider.getReceiverType())) {
            checker.report(Result.failure("override.receiver.invalid",
                    overrider.toString(),
                    enclosingType.toString(),
                    overriddenMethod.toString(),
                    overriddenType.toString(),
                    overrider.getReceiverType(),
                    overriddenMethod.getReceiverType()),
                    overriderTree);
            result = false;
        }
        return result;
    }

    /**
     * Checks the validity of an assignment (or pseudo-assignment) from a value
     * to a variable and emits an error message (through the compiler's
     * messaging interface) if it does.
     * 
     * @param varTree the AST node for the variable
     * @param valueExp the AST node for the value
     * @param errorKey the error message to use if the check fails
     * @param p a checker-specified parameter
     */
    protected void commonAssignmentCheck(Tree varTree, ExpressionTree valueExp, String errorKey, P p) {
        AnnotatedTypeMirror var = factory.getAnnotatedType(varTree);
        assert var != null;
        commonAssignmentCheck(var, valueExp, errorKey, p);
    }

    /**
     * Checks the validity of an assignment (or pseudo-assignment) from a value
     * to a variable and emits an error message (through the compiler's
     * messaging interface) if it does.
     * 
     * @param varType the annotated type of the variable
     * @param valueExp the AST node for the value
     * @param errorKey the error message to use if the check fails
     * @param p a checker-specified parameter
     */
    protected void commonAssignmentCheck(AnnotatedTypeMirror varType, ExpressionTree valueExp, String errorKey, P p) {
        AnnotatedTypeMirror valueType = factory.getAnnotatedType(valueExp);
        assert valueType != null;
        commonAssignmentCheck(varType, valueType, valueExp, errorKey, p);
    }

    /**
     * Checks the validity of an assignment (or pseudo-assignment) from a value
     * to a variable and emits an error message (through the compiler's
     * messaging interface) if it does.
     * 
     * @param varType the annotated type of the variable
     * @param valueType the annotated type of the value
     * @param valueTree the location to use when reporting the error message
     * @param errorKey the error message to use if the check fails
     * @param p a checker-specified parameter
     */
    protected void commonAssignmentCheck(AnnotatedTypeMirror varType,
            AnnotatedTypeMirror valueType, ExpressionTree valueTree, String errorKey, P p) {
        
        boolean success = checker.isSubtype(varType, valueType);
        
        if (options.containsKey("showchecks")) {
            long valuePos = positions.getStartPosition(root, valueTree);
            System.out.printf(
                    " %s (line %3d): %s %s\n     actual: %s %s\n   expected: %s %s\n",
                    (success ? "success" : "FAILURE"),
                    root.getLineMap().getLineNumber(valuePos),
                    valueTree.getKind(), valueTree,
                    valueType.getKind(), valueType,
                    varType.getKind(), varType);
        }
        
        String useKey = (errorKey == null ? "type.incompatible" : errorKey);
        if (!success) {
            checker.report(Result.failure(useKey,
                    valueType.toString(), varType.toString()), valueTree);
        }
    }
}
