package checkers.types;

import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.*;

import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.types.AnnotatedTypeMirror.AnnotatedExecutableType;
import checkers.util.*;
import static checkers.types.AnnotatedTypeMirror.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

import com.sun.tools.javac.code.Symbol;
import com.sun.tools.javac.tree.*;

import static com.sun.tools.javac.tree.JCTree.*;

/**
 * Determines the annotated type (as an {@link AnnotatedTypeMirror}) of an
 * element or AST node.
 * 
 * @manual #writing-a-checker How to write a checker plugin
 */
public class AnnotatedTypeFactory {

    /** The {@link Trees} instance to use for tree node pathfinding. */
    protected final Trees trees;

    /** The root of the syntax tree that this factory operates on. */
    protected final CompilationUnitTree root;

    /** The processing environment to use for accessing compiler internals. */
    protected final ProcessingEnvironment env;

    /** The factory to use for creating annotations. */
    protected final AnnotationFactory annotations;

    /** Utility class for working with {@link Element}s. */
    protected final Elements elements;

    /** Utility class for working with {@link TypeMirror}s. */
    protected final Types types;

    /** Utility class for manipulating annotated types. */
    protected final AnnotatedTypes atypes;

    /** Local utility class for obtaining annotated types from types. */
    protected final TypeCreator creator;

    /** Local utility class for adding annotations to types. */
    protected final TypeAnnotator treeAnnotator;

    /** Local utility class for adding annotations to elements. */
    protected final SimpleElementAnnotator simpleElementAnnotator;

    /**
     * Constructs a factory from the given {@link ProcessingEnvironment}
     * instance and syntax tree root. (These parameters are required so that
     * the factory may conduct the appropriate annotation-gathering analyses on
     * certain tree types.)
     *
     * @param env the {@link ProcessingEnvironment} instance to use
     * @param root the root of the syntax tree that this factory produces
     *            annotated types for
     * @throws IllegalArgumentException if either argument is {@code null}
     */
    public AnnotatedTypeFactory(ProcessingEnvironment env, CompilationUnitTree
            root) {

        if (env == null)
            throw new IllegalArgumentException("null env");

        if (root == null)
            throw new IllegalArgumentException("null root");

        this.env = env;
        this.root = root;
        this.trees = Trees.instance(env);
        this.annotations = new AnnotationFactory(env);
        this.elements = env.getElementUtils();
        this.types = env.getTypeUtils();
        this.atypes = new AnnotatedTypes(env, this);
        this.creator = new TypeCreator();
        this.treeAnnotator = new TypeAnnotator();
        this.simpleElementAnnotator = new SimpleElementAnnotator();
    }
    
    /**
     * Ensures that a type has been constructed properly.
     * 
     * @param type the type to check
     */
    private void checkRep(AnnotatedTypeMirror type) {
        new AnnotatedTypeScanner<Void, Void>() {
            @Override
            public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
                assert type.getElement() != null;
                assert type.getElement() instanceof TypeElement;
                return super.visitDeclared(type, p);
            }

            @Override
            public Void visitExecutable(AnnotatedExecutableType type, Void p) {
                assert type.getElement() != null;
                assert type.getElement() instanceof ExecutableElement;
                return super.visitExecutable(type, p);
            }
            
        }.visit(type);
    }
    
    /**
     * Determines the annotated type of an AST node.
     * 
     * @param tree the AST node
     * @return the annotated type of {@code tree}
     */
    public AnnotatedTypeMirror getAnnotatedType(Tree tree) {

        if (tree == null)
            throw new IllegalArgumentException("null tree");
        
        // Throw an exception for trees that don't have an annotated type.
        switch (tree.getKind()) {
        case COMPILATION_UNIT:
        case IMPORT:
            throw new IllegalArgumentException("invalid tree type: " + tree.getKind());
        default: // continue
        }
        
        AnnotatedTypeMirror type = treeAnnotator.visit(tree, null);
        if (type instanceof AnnotatedDeclaredType) {
            AnnotatedDeclaredType dt = (AnnotatedDeclaredType)type;
            annotateJSR175(dt, dt.getElement());
        }
        checkRep(type);
        return type;
    }

    /**
     * Determines the annotated type of an element.
     * 
     * @param elt the element
     * @return the annotated type of {@code elt}
     */
    public AnnotatedTypeMirror getAnnotatedType(Element elt) {

        if (elt == null)
            throw new IllegalArgumentException("null element");
        
        // Throw an exception for elements with kind "OTHER".
        else if (elt.getKind() == ElementKind.OTHER)
            throw new IllegalArgumentException("invalid element type for "
                    + elt + ": " + elt.getKind());
        
        // We'll get package elements when analyzing fully-qualified accesses
        // (such as "java.util.Arrays.toString"), but we don't care about the
        // annotations on them so we'll just return an empty annotated type.            
        else if (elt.getKind() == ElementKind.PACKAGE)
            return creator.visit(elt.asType());
        
        
        // If there's a tree associated with the element, use it. Otherwise, get
        // the type from the element directly.
        Tree decl = declarationFromElement(elt);
        if (decl != null)
            return getAnnotatedType(decl);

        // Get the annotated type and add JSR 175 annotations from the element.
        AnnotatedTypeMirror type = create(elt.asType(), elt);
        type = annotateJSR175(type, elt);
        
        // There's brokenness with type parameters. Here's why:
        // 
        // Type parameters are identifiers, but we can't get a declaration for
        // them. So we ask for the type of their element, but we can't get a
        // declaration from that either. 

        if (elt instanceof TypeParameterElement) {
            Element parent = ((TypeParameterElement)elt).getGenericElement();
            if (parent != null && parent instanceof TypeElement) {
                TypeElement gen = (TypeElement)parent;
                AnnotatedDeclaredType dt =
                    (AnnotatedDeclaredType)getAnnotatedType(gen);
                int idx = gen.getTypeParameters().indexOf(elt);
                if (dt.getTypeArguments().size() > idx && idx > -1) {
                    AnnotatedTypeMirror result =
                        dt.getTypeArguments().get(idx);
                    if (result instanceof AnnotatedTypeVariable) {
                        // Apply annotations from the upper bound to the type
                        // itself... a job for complete?  (possibly TODO:
                        // similar to the job for moving from an exec. type to
                        // return type?)
                        AnnotatedTypeVariable var = (AnnotatedTypeVariable)result;
                        result.addAnnotations(var.getUpperBound().getAnnotations());
                    }
                    return result;
                }
            }
        }
        checkRep(type);
        return type;
    }
    
    protected AnnotatedTypeMirror create(TypeMirror type) {
        return create(type, null);
    }

    protected AnnotatedTypeMirror create(TypeMirror type, Element elt) {
        creator.reset();
        return creator.visit(type, elt); 
    }
    
    // Takes a TypeMirror, returns an empty (but complete) annotated type.
    protected class TypeCreator extends javax.lang.model.util.AbstractTypeVisitor6<AnnotatedTypeMirror, Element> {

        private List<TypeMirror> visited = new LinkedList<TypeMirror>();

        public void reset() {
            this.visited.clear();
        }

        @Override
        public AnnotatedArrayType visitArray(ArrayType t, Element p) {
            AnnotatedArrayType at = new AnnotatedArrayType(t, env, AnnotatedTypeFactory.this);
            at.setComponentType(visit(t.getComponentType()));
            return at;
        }
        
        @Override
        public AnnotatedTypeMirror visitDeclared(DeclaredType t, Element p) {

            visited.add(t);
                       
            AnnotatedDeclaredType at = new AnnotatedDeclaredType(t, env,
                    AnnotatedTypeFactory.this);

            // Get the element for this type.
            TypeElement elt = (TypeElement) t.asElement();
            
            // Set type arguments.
            List<AnnotatedTypeMirror> args = new LinkedList<AnnotatedTypeMirror>();
            for (TypeMirror arg : t.getTypeArguments())
                args.add(visit(arg));
            at.setTypeArguments(args);
            
            // Set the element.
            // TODO: document when this is set
            at.setElement(elt);
            
            // Defer setting the superclass/interfaces for the completer.
            // TODO: this might be bad for asMemberOf
            
            return at;
        }

        @Override
        public AnnotatedTypeMirror visitError(ErrorType arg0, Element arg1) {
            // There is no AnnotatedError type.
            return null;
        }
        
        @Override
        public AnnotatedExecutableType visitExecutable(ExecutableType t, Element p) {
            
            AnnotatedExecutableType at = new AnnotatedExecutableType(t, env, AnnotatedTypeFactory.this);

            assert p != null;

            // Set the parameter types.
            List<AnnotatedTypeMirror> params = new LinkedList<AnnotatedTypeMirror>();
            for (TypeMirror param : t.getParameterTypes())
                params.add(visit(param));
            at.setParameterTypes(params);

            // Set the return type.
            at.setReturnType(visit(t.getReturnType())); 

            // Set the receiver type.
            at.setReceiverType((AnnotatedDeclaredType)visit(p.getEnclosingElement().asType()));
            
            // Set the thrown type.
            List<AnnotatedTypeMirror> thrown = new LinkedList<AnnotatedTypeMirror>();
            for (TypeMirror thr : t.getThrownTypes())
                thrown.add(visit(thr));
            at.setThrownTypes(thrown);

            // Set the type variables.
            List<AnnotatedTypeVariable> typeVars = new LinkedList<AnnotatedTypeVariable>();
            for (TypeVariable tv : t.getTypeVariables())
                typeVars.add((AnnotatedTypeVariable)visit(tv));
            at.setTypeVariables(typeVars);
            
            // Set the element.
            // TODO: document when this is set
            at.setElement(p);
            
            return at;
        }

        @Override
        public AnnotatedNoType visitNoType(NoType t, Element p) {
            return new AnnotatedNoType(t, env, AnnotatedTypeFactory.this);
        }

        @Override
        public AnnotatedNullType visitNull(NullType t, Element p) {
            return new AnnotatedNullType(t, env, AnnotatedTypeFactory.this);
        }

        @Override
        public AnnotatedPrimitiveType visitPrimitive(PrimitiveType t, Element p) {
            return new AnnotatedPrimitiveType(t, env, AnnotatedTypeFactory.this);
        }

        @Override
        public AnnotatedTypeVariable visitTypeVariable(TypeVariable t, Element p) {
            
            AnnotatedTypeVariable at = new AnnotatedTypeVariable(t, env, AnnotatedTypeFactory.this);

            at.setLowerBound(visit(t.getLowerBound()));

            for (TypeMirror v : visited) {
                if (types.isSameType(t.getUpperBound(), v)) {

                    return at;
                }
            }
            at.setUpperBound(visit(t.getUpperBound()));

            return at;
        }
  
        @Override
        public AnnotatedWildcardType visitWildcard(WildcardType t, Element p) {
            AnnotatedWildcardType at = new AnnotatedWildcardType(t, env, AnnotatedTypeFactory.this);
            if (t.getExtendsBound() != null)
                at.setExtendsBound(visit(t.getExtendsBound()));
            if (t.getSuperBound() != null)
                at.setSuperBound(visit(t.getSuperBound()));
            return at;
        }
    }

    protected class TypeAnnotator extends
        SimpleTreeVisitor<AnnotatedTypeMirror, Void> {

        /**
         * Return the {@link AnnotatedTypeMirror} corresponding to the
         * underlying type of a particular tree.
         * 
         * @param node the node to get an {@link AnnotatedTypeMirror} from
         * @return an {@link AnnotatedTypeMirror} for the given node
         */
        private AnnotatedTypeMirror create(Tree node) {
            return create(node, null);
        }
        
        /**
         * Return the {@link AnnotatedTypeMirror} corresponding to the
         * underlying type of a particular tree, possibly requiring the use of
         * an {@link Element} for completion of the type.
         * 
         * <p>
         * 
         * {@link Element}s are typically only required when creating declared
         * or executable types.
         * 
         * @param node the node to get an {@link AnnotatedTypeMirror} from
         * @param elt the {@link Element} to use to create the type
         * @return an {@link AnnotatedTypeMirror} for the given node
         */
        private AnnotatedTypeMirror create(Tree node, Element elt) {
            
            TypeMirror underlyingType;
            
            // FIXME: this is a kludgy three-step process; it really only needs one step
            
            // 1: Try to get the element from the tree, and use the element to
            // obtain the type.
            Element treeElt = TreeInfo.symbolFor((JCTree)node);
            if (treeElt != null)
                underlyingType = treeElt.asType();
            
            // 2: Obtain the type directly from the tree
            else if (((JCTree) node).type != null)
                underlyingType = ((JCTree) node).type;
            
            // 3: Obtain the type from the tree path
            else {
                TreePath thisPath = trees.getPath(root, node);
                assert thisPath != null : "no path from " + node;
                underlyingType = trees.getTypeMirror(thisPath);
            }
            
            assert underlyingType != null : "no type from " + node;
            
            // Convert it to a an AnnotatedTypeMirror; the result will contain
            // no annotations but its fields (supertypes, bounds, etc.) will
            // be populated.
            creator.reset();
            return creator.visit(underlyingType, elt);
        }
        
        @Override
        public AnnotatedTypeMirror defaultAction(Tree node, Void p) {
            return create(node);
        }
        
        /*
         * Types: add the annotations on the tree, descend into inner types
         */
        
        @Override
        public AnnotatedTypeMirror visitAnnotatedType(AnnotatedTypeTree node,
                Void p) {

            // Get the type of the underlying type.
            AnnotatedTypeMirror type = visit(node.getUnderlyingType(), p);
            assert type != null : "no type from underlying type of " + node;
            
            // Apply the annotations here to the result.
            type.addAnnotations(annotationsFromTree(node));

            return type;
        }

        @Override
        public AnnotatedTypeMirror visitArrayType(ArrayTypeTree node, Void p) {
            
            // Create the base type.
            AnnotatedTypeMirror type = create(node);
            
            // Populate the result.
            AnnotatedTypeMirror component = visit(node.getType(), p);
            ((AnnotatedArrayType)type).setComponentType(component);
            
            return type;
        }

        @Override
        public AnnotatedTypeMirror visitParameterizedType(ParameterizedTypeTree
                node, Void p) {

            // Create the base type.
            AnnotatedTypeMirror type = create(node);
            
            // Populate the result.
            List<AnnotatedTypeMirror> args = new LinkedList<AnnotatedTypeMirror>();
            for (Tree t : node.getTypeArguments()) {
                AnnotatedTypeMirror arg = visit(t, p);
                args.add(arg);
            }
            ((AnnotatedDeclaredType)type).setTypeArguments(args);
            
            return type;
        }

        @Override
        public AnnotatedTypeMirror visitTypeParameter(TypeParameterTree node, Void p) {
            AnnotatedTypeMirror type = create(node);

            List<AnnotatedTypeMirror> bounds = new LinkedList<AnnotatedTypeMirror>();
            for (Tree t : node.getBounds()) {
                AnnotatedTypeMirror bound = visit(t, p);
                bounds.add(bound);
            }

            assert bounds.size() <= 1;
            if (!bounds.isEmpty())
                ((AnnotatedTypeVariable)type).setUpperBound(bounds.get(0));

            return type;
        }

        /*
         * "Type expressions": get the type of the expression's type component
         */
        
        @Override
        public AnnotatedTypeMirror visitTypeCast(TypeCastTree node, Void p) {
            return visit(node.getType(), p);
        }

        @Override
        public AnnotatedTypeMirror visitNewClass(NewClassTree node, Void p) {
            return visit(node.getIdentifier(), p);
        }

        @Override
        public AnnotatedTypeMirror visitNewArray(NewArrayTree node, Void p) {
            
            // For some array creation expressions (such as literal array
            // creation), node.getType() == null and attempts to get the
            // array type ultimately fail.
            // FIXME: this is a hack
            if (node.getType() == null) {
                if (node.getInitializers().size() > 0) {
                    AnnotatedTypeMirror result = visit(node.getInitializers().get(0), p);
                    // TODO: a AnnotatedTypes.getArrayType(componentType) would be nice
                    if (result.getUnderlyingType() != null) {
                        creator.reset();
                        return creator.visit(types.getArrayType(result.getUnderlyingType()));
                    }
                }
            }
            
            AnnotatedTypeMirror type = visit(node.getType(), p);
            AnnotatedTypeMirror result = defaultAction(node, p);
            
            // Move annotations up.
            result.addAnnotations(type.getAnnotations());
            
            assert result instanceof AnnotatedArrayType;
            ((AnnotatedArrayType) result).getComponentType().addAnnotations(
                    annotationsFromArrayCreation(node));
            
            return result;
        }

        /*
         * Uses of types ... 
         */

        // TODO: note that MemberSelect and Identifier should always be similar
        // (if not identical?)

        @Override
        public AnnotatedTypeMirror visitMemberSelect(MemberSelectTree node, Void p) {

            Element elt = elementFromUse(node);

            AnnotatedTypeMirror t = getAnnotatedType(node.getExpression());
            if (t instanceof AnnotatedDeclaredType)
                return atypes.asMemberOf(t, elt); 
            return getAnnotatedType(elt);
        }

        @Override
        public AnnotatedTypeMirror visitIdentifier(IdentifierTree node, Void p) {
                       
            Element elt = elementFromUse(node);
            return getAnnotatedType(elt);
        }

        @Override
        public AnnotatedTypeMirror visitMethodInvocation(MethodInvocationTree
                node, Void p) {

            ExecutableElement elt = elementFromUse(node);
            assert elt != null : "no element from method";
            
            AnnotatedExecutableType methodType = (AnnotatedExecutableType) getAnnotatedType(elt);                       
            
            if (node.getMethodSelect() instanceof MemberSelectTree) {

                // Get the type of the method's receiver for this call.
                MemberSelectTree mst = (MemberSelectTree)node.getMethodSelect();
                AnnotatedTypeMirror recvType = getAnnotatedType(mst.getExpression());
                
                
                // If it's not a declared type, just return the method's return type.
                // FIXME: it should probably always be an AnnotatedDeclaredType
                // In some cases (e.g., "<T> T foo(T bar)") it's a type var.
                if (!(recvType instanceof AnnotatedDeclaredType))
                    return methodType.getReturnType();
                
                AnnotatedDeclaredType t =
                    (AnnotatedDeclaredType)getAnnotatedType(mst.getExpression());

                // Get the type (with substitution);
                AnnotatedTypeMirror type = atypes.asMemberOf(t, elt);

                // Add the JSR 175 method annotations.
                // TODO: this should be done elsewhere? (by asMemberOf abstraction?)
                type = annotateJSR175(type, elt);
                
                return ((AnnotatedExecutableType)type).getReturnType();
            }
            
            return methodType.getReturnType();
        }
        
        /*
         * Declarations ...
         * These often get called by getAnnotatedType(Element).
         */
        
        @Override
        public AnnotatedTypeMirror visitClass(ClassTree node, Void p) {
            TypeElement elt = elementFromDeclaration(node);
            AnnotatedTypeMirror type = create(node, elt);

            List<AnnotatedTypeMirror> typeParams = new LinkedList<AnnotatedTypeMirror>();
            for (TypeParameterTree tp : node.getTypeParameters())
                typeParams.add(visit(tp, p));
            ((AnnotatedDeclaredType)type).setTypeArguments(typeParams);

            return annotateJSR175(type, elt);
        }
        
        @Override
        public AnnotatedTypeMirror visitMethod(MethodTree node, Void p) {
            
            assert node.getReceiver() != null;
            
            ExecutableElement elt = elementFromDeclaration(node);
            
            AnnotatedTypeMirror type = create(node, elt);
            assert type instanceof AnnotatedExecutableType;
            
            // Annotate the receiver type.
            AnnotatedExecutableType methodType = (AnnotatedExecutableType)type;
            methodType.getReceiverType().addAnnotations(
                    annotationsFromTree(node.getReceiver()));
                        
            // Annotate the return type.
            if (node.getReturnType() != null)
                methodType.setReturnType(visit(node.getReturnType(), p));
            
            // Annotate the parameter types.
            List<AnnotatedTypeMirror> paramTypes = new LinkedList<AnnotatedTypeMirror>();
            for (VariableTree tree : node.getParameters())
                paramTypes.add(visit(tree, p));
            methodType.setParameterTypes(paramTypes);
            
            // FIXME: this doesn't add them to the return type
            return annotateJSR175(type, elt);
        }

        @Override
        public AnnotatedTypeMirror visitVariable(VariableTree node, Void p) {
            VariableElement elt = elementFromDeclaration(node);
            AnnotatedTypeMirror type = visit(node.getType(), p);
            return annotateJSR175(type, elt);
        }
        
        // ...
        
        @Override
        public AnnotatedTypeMirror visitConditionalExpression(
                ConditionalExpressionTree node, Void p) {
            
            AnnotatedTypeMirror trueType = getAnnotatedType(node.getTrueExpression());
            AnnotatedTypeMirror falseType = getAnnotatedType(node.getFalseExpression());
            
            // TODO: deeper type intersection could be checker-dependent
            
            Set<AnnotationMirror> intersection = new HashSet<AnnotationMirror>();
            intersection.addAll(trueType.getAnnotations());
            intersection.retainAll(falseType.getAnnotations());
            
            AnnotatedTypeMirror type = create(node);
            type.addAnnotations(intersection);
            
            return type;
        }
    }
    
    protected AnnotatedTypeMirror annotateJSR175(AnnotatedTypeMirror type, Element elt) {
        return simpleElementAnnotator.visit(elt, type);
    }

    protected class SimpleElementAnnotator extends SimpleElementVisitor6<AnnotatedTypeMirror, AnnotatedTypeMirror> {
        @Override
        protected AnnotatedTypeMirror defaultAction(Element e,
                AnnotatedTypeMirror p) {
            assert p != null;
            p.addAnnotations(e.getAnnotationMirrors());
            return p;
        }

        @Override
        public AnnotatedTypeMirror visitExecutable(ExecutableElement e,
                AnnotatedTypeMirror p) {
            assert p instanceof AnnotatedExecutableType;
            AnnotatedExecutableType t = (AnnotatedExecutableType)p;
            t.getReturnType().addAnnotations(e.getAnnotationMirrors());
            return t;
        }        
    }

    /*
     * TODO: the following methods should probably be moved to a util class
     */
    
    /**
     * Gets the element for a class corresponding to a declaration.
     * 
     * @param node 
     * @return the element for the given class
     */
    public final TypeElement elementFromDeclaration(ClassTree node) {
        assert node != null : "null node";
        TreePath path = trees.getPath(root, node);
        if (path == null)
            return (TypeElement)TreeInfo.symbolFor((JCTree)node);
        return (TypeElement)trees.getElement(path);
    }
    
    /**
     * Gets the element for a method corresponding to a declaration.
     * 
     * @param node
     * @return the element for the given method
     */
    public final ExecutableElement elementFromDeclaration(MethodTree node) {
        assert node != null : "null node";
        TreePath path = trees.getPath(root, node);
        if (path == null)
            return (ExecutableElement)TreeInfo.symbolFor((JCTree)node);
        return (ExecutableElement)trees.getElement(path);
    }
    
    /**
     * Gets the element for a variable corresponding to its declaration.
     * 
     * @param node
     * @return the element for the given variable
     */
    public final VariableElement elementFromDeclaration(VariableTree node) {
        assert node != null : "null node";
        TreePath path = trees.getPath(root, node);
        if (path == null)
            return (VariableElement)TreeInfo.symbolFor((JCTree)node);
        return (VariableElement)trees.getElement(path);
    }
    
    /**
     * Gets the element for the method corresponding to this invocation. To get
     * the element for a method declaration, use {@link
     * Trees#getElement(TreePath)} instead.
     *
     * @param node the method invocation
     * @return the element for the method that corresponds to this invocation
     */
    public final ExecutableElement elementFromUse(MethodInvocationTree node) {
        return (ExecutableElement)TreeInfo.symbol((JCTree)node.getMethodSelect());
    }

    /**
     * Gets the element for the declaration corresponding to this identifier.
     * To get the element for a declaration, use {@link
     * Trees#getElement(TreePath)} instead.
     *
     * @param node the identifier
     * @return the element for the declaration that corresponds to this
     * identifier
     */
    public final Element elementFromUse(IdentifierTree node) {
        return TreeInfo.symbol((JCTree)node);
    }

    /**
     * Gets the element for the declaration corresponding to this member
     * access.  To get the element for a declaration, use {@link
     * Trees#getElement(TreePath)} instead.
     *
     * @param node the member access
     * @return the element for the declaration that corresponds to this member
     * access
     */
    public final Element elementFromUse(MemberSelectTree node) {
        return TreeInfo.symbol((JCTree)node);
    }

    // ...
    
    protected final List<? extends AnnotationMirror> annotationsFromTree(AnnotatedTypeTree node) {
        return ((JCAnnotatedType)node).typeAnnotations.annotations;
    }
    
    // TODO: this only works for one dimension
    protected final List<? extends AnnotationMirror> annotationsFromArrayCreation(NewArrayTree node) {
        if (((JCNewArray) node).dimTypeAnnotations.length() > 0)
            return ((JCNewArray) node).dimTypeAnnotations.get(0).annotations;
        return Collections.emptyList();
    }
    
    protected final Tree declarationFromElement(Element elt) {
        // TODO: handle type parameter declarations?
        Tree fromElt = trees.getTree(elt);
        if (fromElt != null)
            return fromElt;
        return TreeInfo.declarationFor((Symbol)elt, (JCTree)root);
    }

    protected static Tree skipParens(Tree tree) {
        assert tree != null : "null tree";
        if (tree.getKind() == Tree.Kind.PARENTHESIZED)
            return ((ParenthesizedTree)tree).getExpression();
        return tree;
    }
}
